/* $Id: kisstnc.cpp 411 2010-05-06 19:29:33Z oh2gve $
 * 
 * Copyright 2007, 2008, 2009, 2010 Tapio Aaltonen
 * 
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// own stuff.
#include "cfg.h"
using Cfg::ConnectionState;
#include "aprspacket.h"
#include "kisstnc.h"

// std stuff.
#include <string>
using std::string;
#include <vector>
using std::vector;
#include <bitset>
using std::bitset;
#include <sstream>
using std::ostringstream;

// wx stuff.
#include <wx/log.h>
#include <wx/utils.h>
#include <wx/string.h>
#include <wx/thread.h>


/// KISS frame start/end marker.
char const FEND = 0xc0;


// Constructor opens serial port by calling constructor of Serial.
KissTnc::KissTnc(wxString const& device, ConnectionState const& state, Cfg::SerialCfg const& portcfg, wxString const& lockdir):
Serial(device, lockdir, state, portcfg),
device_(device), inputBuffer_(), kissFrame_(), tnc2Packet_()
{
}

KissTnc::~KissTnc()
{
}


// PUBLIC MEMBERS


// This is what KissTnc thread keeps doing.
wxThread::ExitCode KissTnc::Entry()
{
	threadStarted();
	
	// This allows main thread to continue with waitForConnect().
	startupDone();
	
	// We start here when main thread says ->Run() and stop the loop when it
	// says ->Delete() or an error happens.
	while ( portIsConnected() && !TestDestroy() )
	{
		// Check for incoming bytes.
		string tmpdata;
		if ( !readDevice(tmpdata) )
		{
			if ( inputEnabled() )
			{
				wxLogError(wxT(" serialkisstnc(") + device_ + wxT("): failed to read data, closing device."));
			}
			else
			{
				wxLogError(wxT(" serialkisstnc(") + device_ + wxT("): Failed to flush device's input buffer, closing device."));
			}
			break;
		}
		
		// Save and process the data if necessary.
		if ( inputEnabled() )
		{
			inputBuffer_ += tmpdata;

			// Look for packet.
			if ( inputBuffer_.size() > 0 )
			{
				// Look for complete KISS frames in our input buffer.
				string frame;
				if ( findKissFrame(frame) )
				{
					// Found a frame, get the packet from it and send to main.
					inputPush( AprsPacket::fromKissFrame(frame) );
				}
			}
		}
		
		// Let's see if we've got outgoing packets.
		AprsPacket const* outpacket = outputPopWait(10);
		if ( outpacket )
		{
			// Process the packet if we have to.
			if ( outputEnabled() )
			{
				if ( !writeToDevice(outpacket->getKissFrame()) )
				{
					wxLogMessage(wxT(" serialkisstnc(") + device_ + wxT("): closing device."));
					delete outpacket;
					break;
				}
			}
			delete outpacket;
		}
	}
	
	// Tell main that we're going down.
	gotDisconnected();
	
	// This thread terminates here.
	return NULL;
}



// PRIVATE


// Find KISS frame from device input buffer.
bool KissTnc::findKissFrame(string& frame)
{
	// Look for end of a frame.
	string::size_type fendPos = inputBuffer_.find(FEND);
	if ( fendPos != string::npos  )
	{
		// Check for zero-lenght frame.
		if ( fendPos == 0 )
		{
			inputBuffer_ = inputBuffer_.erase(0, 1);
			return false;
		}
		
		// Save the frame and remove it from buffer.
		frame = inputBuffer_.substr(0, fendPos);
		inputBuffer_ = inputBuffer_.erase(0, fendPos+1);
		
		return true;
	}
	
	return false;
}
